@@ -19,8 +19,19 @@ from .exceptions import (
 LOGGER = getLogger(name=__name__)
 
 
-class ProtocolError(Exception):
-    'Raised when there is an error during the SMTP conversation.'
+class _ProtocolError(Exception):
+    """
+    Raised when there is an error during the SMTP conversation.
+    Used only internally.
+    """
+
+    def __init__(self, command: str, code: int, message: bytes):
+        self.command = command
+        self.code = code
+        self.message = message.decode(errors='ignore')
+
+    def __str__(self):
+        return f'{self.code} {self.message} (in reply to {self.command})'
 
 
 def _get_mx_records(domain: str, timeout: int) -> list:
@@ -59,8 +70,7 @@ def _smtp_ehlo_tls(smtp: SMTP, helo_host: str):
     code, message = smtp.ehlo(name=helo_host)
     if code >= 300:
         # EHLO bails out, no further SMTP commands are acceptable
-        message = message.decode(errors='ignore')
-        raise ProtocolError(f'EHLO failed: {message}')
+        raise _ProtocolError('EHLO', code, message)
     try:
         smtp.starttls()
         code, message = smtp.ehlo(name=helo_host)
@@ -77,37 +87,34 @@ def _smtp_mail(smtp: SMTP, from_address: EmailAddress):
     code, message = smtp.mail(sender=from_address.ace)
     if code >= 300:
         # MAIL FROM bails out, no further SMTP commands are acceptable
-        message = message.decode(errors='ignore')
-        raise ProtocolError(f'MAIL FROM failed: {message}')
+        raise _ProtocolError('MAIL FROM', code, message)
 
 
 def _smtp_converse(
     mx_record: str, smtp_timeout: int, debug: bool, helo_host: str,
     from_address: EmailAddress, email_address: EmailAddress
-) -> Optional[Tuple[int, str]]:
+):
     """
     Do the `SMTP` conversation, handle errors in the caller.
 
-    Return a `tuple(code, message)` when ambigious, or raise
-    `ProtocolError` on error, and `StopIteration` if the conversation
-    points out an existing email.
+    Raise `_ProtocolError` on error, and `StopIteration` if the
+    conversation points out an existing email.
     """
     if debug:
         LOGGER.debug(msg=f'Trying {mx_record} ...')
-    smtp = SMTP(timeout=smtp_timeout, host=mx_record)
-    smtp.set_debuglevel(debuglevel=2 if debug else False)
-    _smtp_ehlo_tls(smtp=smtp, helo_host=helo_host)
-    _smtp_mail(smtp=smtp, from_address=from_address)
-    code, message = smtp.rcpt(recip=email_address.ace)
-    smtp.quit()
-    if code == 250:
-        raise StopIteration
-    elif code >= 500:
-        message = message.decode(errors='ignore')
-        raise ProtocolError(f'RCPT TO failed: {message}')
-    # Ambigious return code, can be graylist, temporary problems,
-    # quota or mailsystem error
-    return code, message.decode(errors='ignore')
+    with SMTP(timeout=smtp_timeout) as smtp:
+        smtp.set_debuglevel(debuglevel=2 if debug else False)
+        code, message = smtp.connect(host=mx_record)
+        if code >= 400:
+            raise _ProtocolError('connect', code, message)
+        _smtp_ehlo_tls(smtp=smtp, helo_host=helo_host)
+        _smtp_mail(smtp=smtp, from_address=from_address)
+        code, message = smtp.rcpt(recip=email_address.ace)
+        if code == 250:
+            # Address valid, early exit
+            raise StopIteration
+        elif code >= 500:
+            raise _ProtocolError('RCPT TO', code, message)
 
 
 def _check_one_mx(
@@ -119,27 +126,25 @@ def _check_one_mx(
     `StopIteration` if this MX accepts the email.
     """
     try:
-        result = _smtp_converse(
+        _smtp_converse(
             mx_record=mx_record, smtp_timeout=smtp_timeout, debug=debug,
             helo_host=helo_host, from_address=from_address,
             email_address=email_address)
     except SMTPServerDisconnected:
         return True
-    except (SocketError, ProtocolError) as error:
+    except (SocketError, _ProtocolError) as error:
         error_messages.append(f'{mx_record}: {error}')
         return False
-    if result:
-        error_messages.append(f'{mx_record}: {result[0]} {result[1]}')
-        return True
-    return False
+    return True
 
 
 def _check_mx_records(
-    mx_records: list, smtp_timeout: int, helo_host: str,
-    from_address: EmailAddress, email_address: EmailAddress,
-    debug: bool,
-) -> Optional[bool]:
+        mx_records: list, smtp_timeout: int, helo_host: str,
+        from_address: EmailAddress, email_address: EmailAddress,
+        debug: bool) -> Optional[bool]:
     'Check the mx records for a given email address.'
+    # TODO: Raise an ambigious exception, containing the messages? Will
+    # be a breaking change.
     error_messages = []
     found_ambigious = False
     for mx_record in mx_records:
@@ -150,11 +155,12 @@ def _check_mx_records(
                 email_address=email_address, smtp_timeout=smtp_timeout,
                 debug=debug)
         except StopIteration:
+            # Address valid, early exit
             return True
     # If any of the mx servers behaved ambigious, return None, otherwise raise
     # an exception containing the collected error messages.
     if not found_ambigious:
-        raise AddressNotDeliverableError(error_messages)
+        raise AddressNotDeliverableError(error_messages=error_messages)
 
 
 def mx_check(
@@ -165,9 +171,8 @@ def mx_check(
 ) -> Optional[bool]:
     """
     Return `True` if the host responds with a deliverable response code,
-    `False` if not-deliverable.
-    Also, return `None` if there if couldn't provide a conclusive result
-    (e.g. temporary errors or graylisting).
+    `False` if not-deliverable. Also, return `None` if there if couldn't
+    provide a conclusive result (e.g. temporary errors or graylisting).
     """
     host = helo_host or gethostname()
     from_address = from_address or email_address
